import { JsonObject } from 'type-fest';

type SafeEntity$1<Entity extends JsonObject, Components extends keyof Entity> = Entity & Required<Pick<Entity, Components>>;
/**
 * Container for Entities
 */
declare class World<Entity extends JsonObject> {
    #private;
    get archetypes(): Set<Archetype<Entity, Array<keyof Entity>>>;
    get entities(): ReadonlySet<Entity>;
    clearEntities(): void;
    archetype<Components extends Array<keyof Entity>>(...components: Components): Archetype<SafeEntity$1<Entity, (typeof components)[number]>, typeof components>;
    createEntity(): Entity;
    /**
     * Create an entity with the given components. This is a type-safe version
     * __but__ it is of a point in time. When the entity is created. So don't
     * rely on it to be type-safe in the future when used within systems.
     */
    createEntity<T extends Entity>(entity: T): keyof typeof entity extends never ? never : Pick<Entity & T, keyof typeof entity>;
    deleteEntity(entity: Entity): boolean;
    addEntityComponents<T extends Entity, Component extends keyof Entity>(entity: T, component: Component, value: NonNullable<Entity[Component]>): T & Record<typeof component, typeof value>;
    removeEntityComponents<T extends Entity, Component extends keyof Entity>(entity: T, ...components: Array<Component>): void;
}

type SafeEntity<Entity extends JsonObject, Components extends keyof Entity> = Entity & Required<Pick<Entity, Components>>;
/**
 * An archetype is a collection of entities that share the same components.
 * Archetypes should not be constructed directly, but rather through the
 * `World` class using the `archetype` method.
 */
declare class Archetype<Entity extends JsonObject, Components extends Array<keyof Entity>> {
    #private;
    constructor({ entities, world, components, without, }: {
        world: World<Entity>;
        entities: Set<Entity>;
        components: Components;
        without?: Array<Exclude<keyof Entity, Components[number]>>;
    });
    get entities(): ReadonlySet<SafeEntity<Entity, Components[number]>>;
    get components(): Readonly<Components>;
    get excluding(): Readonly<Array<Exclude<keyof Entity, Components[number]>>>;
    matches(entity: Entity): boolean;
    addEntity(entity: Entity): Archetype<Entity, Components>;
    removeEntity(entity: Entity): Archetype<Entity, Components>;
    clearEntities(): void;
    /**
     * Returns a new archetype based on the current archetype, but excludes the
     * specified components.
     * @param components Components that should **not** be present on the entity
     * @returns
     */
    without<Component extends keyof Entity>(...components: Component[]): Archetype<SafeEntity<Omit<Entity, (typeof components)[number]>, Exclude<Components[number], (typeof components)[number]>>, Array<Exclude<Components[number], (typeof components)[number]>>>;
}

export { Archetype as A, type SafeEntity$1 as S, World as W };
